__all__ = ["Selector"]

from pyvorse.core import basis
from typing import Callable, List, Any, Hashable
import random
import json

class Selector(basis):
    """Randomized selections from json-like data"""

    SELECT = random.Random(0)
    @classmethod
    def ResetSeed(cls, x=0):
        cls.SELECT.seed(x)
        return cls.SELECT
    
    
    @classmethod
    def weighted(cls, data: List[List]) -> Any:
        "Data format - List[List[]]. List of lists with len == 2. First value - value, Second - weight of this value. Returns randomly selected value"
        summ = 0
        for pair in data:
            if not isinstance(pair, list): raise TypeError("Object must be a list.")
            if not len(pair) == 2: raise ValueError(f"Pair must have length equal to 2, but is {len(pair)}")
            # Calculate weights sum:
            summ += pair[1]
        
        # Generate value between 0 and $summ, step by step inside variations, and on each step subtract weight. When variable goes 0 or less - select pair, and its value
        var = cls.SELECT.randrange(0, summ)
        # select
        for pair in data:
            var -= pair[1]
            if var <= 0: return pair[0]
        raise Exception("Unknown error with weighted selection")
    
    @classmethod
    def weightedExt(cls, data: List[List]) -> Any:
        "Analog of .weighted, but inner lists can contain more then one value, uses last value as weight. Returns slice of list without last value (weight)"
        summ = 0
        for pair in data:
            if not isinstance(pair, list): raise TypeError("Object must be a list.")
            # Calculate weights sum:
            summ += pair[-1]

        # Generate value between 0 and $summ, step by step inside variations, and on each step subtract weight. When variable goes 0 or less - select pair, and its value
        var = cls.SELECT.randrange(0, summ)
        # select
        for pair in data:
            var -= pair[-1]
            if var <= 0:
                return pair[:-1]
        raise Exception("Unknown error with weighted selection")
    
    @classmethod
    def multi(cls, count: int, method: Callable, *args, **kwargs):
        res = []
        for i in range(count):
            res.append(method(*args, **kwargs))
        return res





__all__ = ["deep", "dci", "DeepClass"]

from pyvorse.core import copyAllProps

class deep(dict):
    """Like dict, allows to use prototype hierarchy, and access entries as attribute for get-obly mode. use methodes, with names starts with 'deep'"""
    def __init__(self, *args, **kwargs):
        super(deep, self).__init__(*args, **kwargs)
        self.proto = None # Property, thzat saves link to prototype object
        self.deepinstance = None # used only if object is instance of deep class
    
    def deepget(self, name, *, depth:int = 0):
        if name in self:
            return self[name]
        elif self.proto == None:
            raise KeyError(f"No key '{name}' in object. Depth = {depth}")
        else:
            return self.proto.deepget(name, depth= depth + 1)
    
    def deephas(self, name):
        if name in self:
            return True
        elif self.proto == None:
            return False
        else:
            return self.proto.deephas(name)
    
    def deepfind(self, name):
        "Returns instance in hirerarchy, that has a key, passed as argument. Returns None if nothing found."
        if name in self:
            return self
        elif self.proto == None:
            return None
        else:
            return self.proto.deepfind(name)
    
    def deepset(self, name, value):
        "Allows to set value on key with actual key depth, or on depth 0 if key not found"
        if self.deephas(name):
            self.deepfind[name] = value
            return self
        self[name] = value
    
    def deepdel(self, name):
        "Allows to delete entry on key with actual key depth, or on depth 0 if key not found"
        if self.deephas(name):
            del self.deepfind(name)[name]
            return self
    
    def setproto(self, proto):
        self.proto = proto
        return self
    
    def fill(self, **kwargs):
        for key, value in kwargs.items():
            self.deepset(key, value)
        return self
    
    def up(self):
        return self.proto
    
    def upsafe(self):
        return self.proto if self.proto != None else self
    
    def hasprototype(self):
        return False if self.proto == None else True
    
    def isdeepinstance(self):
        "Tests object to be an instance of some deep class, use .deepinstance to get deepclass object"
        return self.deepinstance != None
    
    def __getattr__(self, name):
        return self.deepget(name)
    
    @classmethod
    def create(cls, proto=None, *origins, **kwargs):
        if not isinstance(proto, deep):
            raise TypeError("Only deep object can be uset as prototype for other deep objects.")
        return cls(*origins).setproto(proto).fill(*kwargs)
    
    def callAsMethod(self, name, *args, **kwargs):
        "Allows calling functions inside deep object, passing deep object as first argument"
        o = self.deepget(name)
        if not callable(o):
            raise TypeError("object exists, but not callable")
        return o(self, *args, **kwargs)
    
    def copy(self):
        "Creates full copy of object"
        o = self.__class__(super(deep, self).copy()) # not an error, super of deep, to prevent erros on cloning deepclasses
        copyAllProps(o, self)
        return o




"""
LEXICON:
prototype - only in classes. Link to object, that will be used as prototype for instancesself.
proto - inside any deep object, link to the prototype of object itself.
"""

class deepclass(deep):
    """Builder for any deep objects with prototype, created deep object wit prototype from deepclass.
    You can extend the deepclass by method .extend(new_prototype). Works like class for"""
    def __init__(self, prototype = None):
        super(deepclass, self).__init__()
        self.prototype = prototype # contains link to proto object for created objects
        self.ensure = None
    
    def setprototype(self, prototype: deep):
        "set deep object, that be set as proto of instances for it."
        self.prototype = prototype
        return self
    
    def setparent(self, parent):
        self.parent = parent
        return self
    
    def ensurefields(self, *names:str):
        self.ensure = tuple(names)
        return self
    
    def extend(self, new_prototype: deep, *names: str):
        dc = deepclass.create(self).setparent(self).setprototype(new_prototype).ensurefields(*names)
        return dc._chainfix()
    
    def _chainfix(self):
        if not self.prototype.proto is self.proto.prototype:
            self.prototype.setproto(self.proto.prototype)
        return self
    
    def new(self, **kwargs):
        o = self.prototype.__class__.create(self.prototype, **kwargs)
        o.deepinstance = self
        for name in self.ensure:
            o[name] = None
        return o


dci = deepclass()



__all__ = ["smartlist", "handle", "objectAddress"]

from typing import Callable
from pyvorse.core import basis, unibox


class smartlist(list):
    "List, with additional methods like foreach, find..."

    def find(self, finder: Callable):
        "uses function with 3 arguments (value, index, _list)\nreturns tuple (item, index) or (None, None)"
        i = 0
        for item in self:
            if finder(item, i, self):
                return item, i
            i += 1
        return None, None
    
    def select(self, finder: Callable):
        "uses function with 3 arguments (value, index, _list)\nreturns value or None"
        i = 0
        for item in self:
            if finder(item, i, self):
                return item
            i += 1
        return None
    
    def filtered(self, finder: Callable):  # iterator
        "Custom iterator, that generates only values filtered by predicate"
        i = 0
        for item in self:
            if finder(item, i, self):
                yield item
            i += 1
    
    def filter(self, finder: Callable):
        "returns new instance of smart list with filtered values from initial smart list"
        return smartlist([item for item in self.filtered(finder)])
    
    def any(self, predicate: Callable):
        "Returns true if at least one of values returns true on predicate call. Predicate(value, index, _list)"
        i = 0
        for item in self:
            if predicate(item, i, self):
                return True
            i += 1
        return False
    
    def all(self, predicate: Callable):
        "Returns true if at all values returns true on predicate call. Predicate(value, index, _list)"
        i = 0
        for item in self:
            if not predicate(item, i, self):
                return False
            i += 1
        return True
    
    def astuple(self):
        return tuple(self)
    
    def asset(self):
        return set(self)
    
    def aslist(self):
        return list(self)
    
    # works
    def foreach(self, func: Callable):
        "Calls function for every element in list. Func(value, index, _list)"
        i = 0
        for item in self:
            func(item, i, self)
            i += 1
    
    def mapper(self, func: Callable):
        "Calls function for every element in list and writs it return to result list. Func(value, index, _list)"
        i = 0
        result = smartlist()
        for item in self:
            result.append(func(item, i, self))
            i += 1
        return result

    def reducer(self, func: Callable, previnit = None):
        "Reduces list to one result value. Func(prev, value, index, _list)"
        i = 0
        prev = previnit
        for item in self:
            prev = func(prev, item, i, self)
            i += 1
        return prev
    
    def sort(self, sorter: Callable, invert:bool=False):
        """Sorter(value1, value2)
        Returns: -1 >>> value1 < value2
        Returns: 0  >>> value1 == value2
        Returns: 1  >>> value1 > value2
        In result - larger moved to th end of list.
        You can use invert argument all.reverse() method
        """
        self._sort(0, len(self) - 1, sorter=sorter)
        if invert: self.reverse()
        return self
    
    def _partiate(self, start, end, *, sorter):
        pivot = self[start]
        low = start + 1
        high = end

        while True:
            # If the current value we're looking at is larger than the pivot
            # it's in the right place (right side of pivot) and we can move left,
            # to the next element.
            # We also need to make sure we haven't surpassed the low pointer, since that
            # indicates we have already moved all the elements to their correct side of the pivot
            while low <= high and sorter(self[high], pivot) > -1:
                high = high - 1

            # Opposite process of the one above
            while low <= high and sorter(self[low], pivot) < 1:
                low = low + 1

            # We either found a value for both high and low that is out of order
            # or low is higher than high, in which case we exit the loop
            if low <= high:
                self[low], self[high] = self[high], self[low]
                # The loop continues
            else:
                # We exit out of the loop
                break

        self[start], self[high] = self[high], self[start]

        return high

    def _sort(self, start, end, *, sorter):
        if start >= end:
            return

        p = self._partiate(start, end, sorter=sorter)
        self._sort(start, p-1, sorter=sorter)
        self._sort(p+1, end, sorter=sorter)
    
    def sliced(self, start, end):
        return smartlist(self[start:end])
    
    def slicedstart(self, start):
        return smartlist(self[start:])
    
    def slicedend(self, end):
        return smartlist(self[:end])


class handle(basis):
    """Line of handlers for an object"""

    def __init__(self):
        super().__init__()
        self.way = []
    
    def __call__(self, callback):
        "callback uses one argument, link to current values."
        self.way.append(callback)
        return self
    
    def use(self, data):
        if isinstance(data, unibox):
            return self._use(data)
        return self._use(unibox(data))

    def _use(self, box):
        i = -1 # 0 on first iteration
        for cb in self.way:
            i += 1
            box["Step"] = i
            if not box.stopped:
                try:
                    cb(box)
                except Exception as e:
                    if not "errors" in box:
                        box["errors"] = []
                    box["errors"].append(e)
        return box


class objectAddress(basis):
    """Saves linear cobination of keys and indexes, and callbacks for any object. Used for deep structures,
    accessed wit [] operator"""
    def __init__(self):
        super(objectAddress, self).__init__()
        self.way = []
    
    def __call__(self, callback):
        "callback uses one argument, link to current values."
        self.way.append(callback)
        return self
    
    def __getitem__(self, key):
        "callback uses one argument, link to current values."
        self.way.append(key)
        return self
    
    def use(self, o):
        tmp = o
        for key in self.way:
            if callable(key):
                tmp = key(tmp)
            else:
                tmp = tmp[key]
        return tmp



__all__ = ["trin", "hexcount"]

def trin(c: int) -> int:
    "returns triangular number of an number"
    return int((c * (c + 1)) / 2)


def hexcount(c: int) -> int:
    "returns count of hexagonal cells by radius, c = 1 means single cell"
    return int(6 * (trin(c) - c) + 1)



from typing import Callable
from pyvorse.core import basis
from pyvorse.tests import isinrange

class DictionaryPipeline(dict):
    "Represents package of function, with access by it`s names. In cell uses item ['target'] for select next. "
    def __call__(self, func: Callable):
        if not hasattr(func, "__name__"):
            return func
        self[func.__name__] = func
        return self
    
    def append(self, funcOrName):
        "Advanced append decorator, that allows rename function"
        if callable(funcOrName) and hasattr(funcOrName, "__name__"):
            self[funcOrName.__name__] = funcOrName
            return funcOrName
        elif isinstance(funcOrName, str):
            def decor(func: Callable):
                self[funcOrName] = func
                return func
            return decor
        raise ValueError("Possible used lambda function without name")
    
    def __getattr__(self, name):
        return self[name]
    
    def defaultpointer(self, cell, smeta):
        if not "target" in cell or not cell['target'] in self:
            return
        return self[cell['target']]


class ListPipeline(list):
    "Represents package of function, with access by it`s names. In cell uses item ['target'] for select next. "

    def __call__(self, func: Callable):
        self.apend(func)
        return self

    def append(self, func):
        "Advanced append decorator, that allows rename function"
        self.apend(func)
        return self

    def __getattr__(self, index):
        return self[index]

    def defaultpointer(self, cell, smeta):
        if not "target" in cell or not isinrange(self, cell['target']):
            return
        return self[cell['target']]
    



__all__ = ['attribute', 'w', 'basis', 'throws', 'isthrows',
           'trycall', 'Operators', 'calloperators', 'functor', 'reprint', "safe", "copyAllProps"]

from typing import Callable
from enum import IntFlag, auto
from functools import wraps
########################################################################

# Allows converting function to an object, that
class attribute:
    "Processor, that allows to call function with name, passed as access to attribute of an object."

    def __init__(self, func):
        if callable(func):
            self.__function__ = func
        else:
            raise ValueError("Object is not callable")

    def __call__(self, name):
        return self.__function__(name)

    def __getattr__(self, name):
        return self.__function__(name)

    def __getitem__(self, name):
        return self.__function__(name)


# w.some_word == "some_word"
@attribute
def w(name: str):
    "Other way to in-code write identifier strings"
    return name

########################################################################

# base class with some modifications
class basis(object):
    "Alternative base class. Rewrite representation with .repr(self, data=None) method"

    def __str__(self, data=None):
        return self.repr(data)

    def repr(self, data=None):
        if data != None:
            return f"[{self.__class__.__qualname__}]: {data}"
        return f"[{self.__class__.__qualname__}]"

    def __repr__(self, data=None):
        return self.repr(data)

    @classmethod
    def new(cls, *args, **kwargs):
        return cls(**kwargs)

    @classmethod
    def getClass(cls):
        return cls

    @classmethod
    def getClassName(cls):
        return cls.__name__

    @classmethod
    def getClassQualName(cls):
        return cls.__qualname__

    def id(self):
        return id(self)

    def hexid(self):
        return hex(id(self))

    def __hash__(self):
        return id(self)

    def getAttribute(self, name: str, default = None):
        "Safely access attribute"
        return getattr(self, name, default)


########################################################################
# Marks function with possible exception classes, us ass callable decorator
def throws(*exceptions):
    """Marks method with exceptions, that can throw
    Callable decorator:
    
    @throws(ValueError, IndexError)
    def function():
        pass
    """
    def editor(func: Callable):
        func.__throws__ = tuple(exceptions)
        return func
    return editor

# For testing before invoking
def isthrows(func: Callable, cls):
    """Tests function to __throws__ property"""
    if not hasattr(func, '__throws__'):
        return False
    return cls in func.__throws__

# alternative for using try: except:
def trycall(func: Callable, *args, **kwargs):
    """Calls function inside try:except:
    use like:
    result, exception = trycall(getattr, o, "attrname")
    """
    try:
        return (func(*args, **kwargs), None)
    except BaseException as e:
        return (None, e)

# Safe wrapper for fuction
def safe(func: Callable):
    """Converts function to version, thst called safely, not throws exceptions.
    Returns tuple with 2 elments:
    - if no error: (result, None)
    - if error: (None, exception_object)
    converted functions can be used as:
    result, exception = safefunc()
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        return trycall(func, *args, **kwargs)
    return wrapper

###########################################################################
# Wrapper for calling function on operators

class Operators(IntFlag):
    Add = auto()
    Subtract = auto()
    Multiply = auto()
    Divide = auto()
    FloorDivide = auto()
    Modulo = auto()
    Power = auto()
    All = Add | Subtract | Multiply | Divide | FloorDivide | Modulo | Power

class OperandHandler(basis):
    def __init__(self, func: Callable, op: Operators):
        self.__operators__ = {}
        self.__oncall__ = func
        self.append(func, op)

    def append(self, func: Callable, op: Operators):
        self.__operators__[op] = func
        return self

    def __add__(self, other):
        for key, value in self.__operators__.items():
            if Operators.Add in key:
                return value(other)
        raise ValueError(f"Operator {Operators.Add.name} not supported")

    def __sub__(self, other):
        for key, value in self.__operators__.items():
            if Operators.Subtract in key:
                return value(other)
        raise ValueError(f"Operator {Operators.Subtract.name} not supported")

    def __mul__(self, other):
        for key, value in self.__operators__.items():
            if Operators.Multiply in key:
                return value(other)
        raise ValueError(f"Operator {Operators.Multiply.name} not supported")

    def __floordiv__(self, other):
        for key, value in self.__operators__.items():
            if Operators.FloorDivide in key:
                return value(other)
        raise ValueError(
            f"Operator {Operators.FloorDivide.name} not supported")

    def __truediv__(self, other):
        for key, value in self.__operators__.items():
            if Operators.Divide in key:
                return value(other)
        raise ValueError(f"Operator {Operators.Divide.name} not supported")

    def __mod__(self, other):
        for key, value in self.__operators__.items():
            if Operators.Modulo in key:
                return value(other)
        raise ValueError(f"Operator {Operators.Modulo.name} not supported")

    def __pow__(self, other):
        for key, value in self.__operators__.items():
            if Operators.Power in key:
                return value(other)
        raise ValueError(f"Operator {Operators.Power.name} not supported")

    def __call__(self, *args, **kwargs):
        return self.__oncall__(*args, **kwargs)

    def editoperands(self, op: Operators):
        def wrapper(func: Callable):
            self.append(func, op)
        return wrapper

def calloperators(op: Operators = Operators.All):
    """Callable decorator, allows function to be called on using it as operand of some operator.exe
    Using:
    @calloperators(Operators.All)
    def reprint(*args, **kwargs):
        print(*args, **kwargs)
        if len(args) == 1:
            return args[0]
    """
    def wrapper(func: Callable):
        return OperandHandler(func, op)
    return wrapper

@calloperators(Operators.All)
def reprint(*args, **kwargs):
    print(*args, **kwargs)
    if len(args) == 1:
        return args[0]

##########################################################################
# Functor wrapper

def functor(func: Callable):
    """Converts function to it`s copy, that will be called with function object as first argument.
    ```py
    @functor
    def bug(functor, arg*):
        # you can use properties of function
        return functor.__name__
    ```
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(wrapper, *args, **kwargs)
    return wrapper

##########################################################################
class unibox(basis):
    """Container for data. Item accessors uses metadata dictionary. Allows stopping.
    On iteratio - key-value pairs of metadata. Attribute get-access - derivied to item inside container.
    Data can be fully accessed by «.inbox» field.
    """

    def __init__(self, o=None, meta={}):
        self.inbox = o.inbox if isinstance(o, self.getClass()) else o
        self.meta = meta
        self.stopped: bool = False

    def __getitem__(self, name):
        return self.meta[name]

    def __setitem__(self, name, value):
        self.meta[name] = value

    def __delitem__(self, name):
        del self.meta[name]

    def __contains__(self, name):
        return self.meta.__contains__(name)

    def get(self, name, default=None):
        try:
            return self.meta[name]
        except (KeyError, NameError, IndexError, AttributeError):
            return default

    def __getattr__(self, name):
        return getattr(self.inbox, name)

    def clearmeta(self):
        self.meta.clear()

    def setval(self, value):
        self.inbox = value
        return self

    def __iter__(self):
        return self.meta.items()

    def stop(self):
        self.stopped = True
        return self

    def unstop(self):
        self.stopped = False
        return self

    def isStopped(self):
        return self.stopped


##########################################################################
# Dangerous
def copyAllProps(dest, source):
    "Forces object to clone all properties, not started from duble underscore '__'"
    dd = source.__dict__
    for key, value in dd.items():
        if not key.startswith('__'):
            setattr(dest, key, value)



__all__ = ["spider"]

from pyvorse.core import basis, unibox
from typing import Callable

class spider(basis):
    """Creates an object with functions that automatically explores complex data structures. 
    Seeker function - to find next element. cb(structure, element, box) -> «next_element». 
    Worker function - to do something with data element.
    Stop box to finish spider walk. """
    def __init__(self, seeker: Callable, worker: Callable):
        self.seeker = seeker
        self.worker = worker
    
    def launch(self, struct):
        dot = struct
        box = unibox()

        while not box.stopped:
            dot = self.seeker(struct, dot, box)
            self.worker(struct, dot, box)
        
        return box
    
    def use(self, struct):
        return self.launch(struct)




__all__ = ["poslist"]

from pyvorse.core import trycall

class poslist(dict):
    """Behaves mostly like default list, but saves positions, allows skipsself. Trigger - value used as «null» for skipping cells"""

    def __init__(self, *values, trigger=None):
        self.next = 0
        if len(values) > 0:
            self.appends(*values, trigger=None)

    def _autoupdate(self):
        self.next = self._getmaxindex() + 1

    def _getmaxindex(self):
        return trycall(max, tuple(self.keys()))[0] or 0
        

    def append(self, value):
        self[self.next] = value
        self._autoupdate()
        return self
    
    def insert(self, value, index: int = None):
        if index == None:
            index = self.next
        self[index] = value
        self._autoupdate()
        return self
    
    def skip(self):
        self.next += 1
    
    def appends(self, *values, trigger=None):
        "Use trigger to mark skip slots."
        for value in values:
            if value == trigger:
                self.skip()
            else:
                self.append(value)
        return self

    
    def __iter__(self):
        for value in self.values():
            yield value
    
    def indexes(self):
        return self.keys();
    
    def getor(self, index, default=None):
        if not index in self:
            return default
        return self[index]
    
    def __setitem__(self, index, value):
        self._autoupdate();
        return super().__setitem__(index, value)
    
    def __delitem__(self, index):
        self._autoupdate()
        return super().__delitem__(index)
    
    def indexes(self):
        return self.keys()
    
    def pop(self, key):
        res = super().pop(key)
        self._autoupdate()
        return res
    
    def popitem(self):
        res = super().popitem()
        self.autoupdate()
        return res



__all__ = ["Table", "Column", "Row"]

from pyvorse.core import basis
from pyvorse.data.positioned import poslist

class Column(basis):
    "Represents headers for tables"
    def __init__(self, name: str, align: int = 0, default = None):
        self.name = name
        self.align = align
        self.default = default
    
    def alignLeft(self):
        self.align = -1
    
    def alignCenter(self):
        self.align = 0
    
    def alignRight(self):
        self.align = 1
    
    def alignSwitch(self):
        self.align *= -1
    
    def setdefault(self, default):
        self.default = default

class Row(poslist, basis):
    "Represents datarow, positioned, based on dict, but behave like list"

class ColumnData(basis):
    "Contains all data about table row"
    def __init__(self, header: Column, *dataset):
        self.header = header
        self.data = dataset
    
    def __getitem__(self, index):
        return self.data[index]

class Table(basis):
    """Contains columns and rows. You can use functions as values and get items by .getsmart, function will be called
    with 3 arguments, link to table object, indexes for row and column. That allows dynamic calculated values, like summary."""
    def __init__(self, *columns: Column):
        super(Table, self).__init__()
        self.columns = [*columns]
        self.rows = []
    
    def cleardata(self):
        self.rows = []
    
    def clearall(self):
        self.cleardata()
        self.columns = []
    
    def append(self, row: Row):
        self.rows.append(row)
        return self
    
    def appends(self, *items, trigger=None):
        self.rows.append(Row(*items, trigger=trigger))
        return self
    
    def addheader(self, header: Column):
        self.columns.append(header)
    
    def adddefaultheader(self, name: str, align: int = 0, default=None):
        self.columns.append(Column(name, align, default))
    
    def getrow(self, index: int):
        return self.rows[index]
    
    def getheader(self, index: int):
        return self.columns[index]
    
    def getcolumn(self, index: int):
        res = []
        for row in self.rows:
            res.append(row.getor(index, self.columns[index].default))
        return ColumnData(self.columns[index], *res)
    
    def getitem(self, row, col):
        return self.rows[row].getor(col, self.columns[col].default)
    
    def setitem(self, row, col, value):
        self.rows[row][col] = value
        return self
    
    def delitem(self, row, col):
        return self.setitem(row, col, self.columns[col].default)
    
    def getdynamic(self, row, col):
        item = self.getitem(row, col)
        return item if not callable(item) else item(self, row, col)

    def __getitem__(self, index):
        return self.getcolumn(index)



__all__ = ["loop", "scope", "stack"]

from pyvorse.core import basis

class loop(list, basis):
    "List, but works with indexes in a different way. Works as looped list. Iterations in cycle ar finite."

    def __getitem__(self, index):
        index = index % len(self)
        return super().__getitem__(index)

    def __setitem__(self, index, value):
        index = index % len(self)
        return super().__setitem__(index, value)

    def __delitem__(self, index):
        index = index % len(self)
        return super().__delitem__(index)

    def insert(self, index, object):
        index = index % len(self)
        return super().insert(index, object)

    def pop(self, index):
        index = index % len(self)
        return super().pop(index)

    def first(self, index: int = 0):
        return self[index]

    def last(self, index: int = 0):
        return self[len(self) - index - 1]
    

class scope(dict, basis):
    "Actualy dict, but extended wit attribute-like access. Iterating by default equivalent to dict.items(). Allows to find value by index «.(index: int)»"

    def __call__(self, index: int):
        return self.getByIndex(index)
    
    def getByIndex(self, index: int):
        return tuple(self.values())[index]
    
    def getKeyByIndex(self, index: int):
        return tuple(self.keys())[index]
    
    def __iter__(self):
        return self.items()
    
    def __getattr__(self, name):
        return self[name]
    
    def __setattr__(self, name, value):
        if name in self:
            self[name] = value
        super().__setattr__(name, value)
    
    def clone(self):
        return scope(super().clone())
    
    def getDynamic(self, key):
        "if value is callable, returns result of calling it function or method, with scope instance as first arg."
        if callable(self[key]):
            return self[key](self)
        return self[key]


class stack(list, basis):
    """Simplest stack realization based on list. Use .push and .pull methods
    You can use validation function, set with .setvalidation(callback(stack, item))
    Full list, but with new methods."""
    validator = None

    def setvalidation(self, callback):
        self.validator = callback
        return self
    
    def validate(self, value) -> bool:
        if callable(self.validator):
            return self.validator(self, value)
        return True

    def push(self, item):
        if self.validate(item):
            return self.append(item)
        raise ValueError(f"Item blocked by validator, use .append to ignore it. // {item}")
    
    def pushes(self, *items):
        for item in items:
            self.push(item)
        return self
    
    def pull(self):
        return self.pop()
    
    def leftpush(self, item):
        if self.validate(item):
            return self.insert(0, item)
        raise ValueError(
            f"Item blocked by validator, use .append to ignore it. // {item}")

    def leftpushes(self, *items):
        for item in items:
            self.leftpush(item)
        return self

    def leftpull(self):
        return self.pop(0)
    
    def repr(self, data=None):
        return " >> ".join(map(str, self))
    
    def empty(self) -> bool:
        return len(self) == 0
    



__all__ = ["hashable", "haskey", "isand", "isor", "eqand", "eqor", "haslen"]

from typing import overload


def hashable(elem):
    """
    Returns True if element is able to hash and vice-versa
    """
    try:
        x = iter(elem)
        next(x)
    except:
        return False
    else:
        return True


def haskey(item, key):
    try:
        item[key]
    except (IndexError, KeyError):
        return False
    else:
        return True


def haslen(item):
    try:
        len(item)
    except (TypeError, AttributeError):
        return False
    else:
        return True


# logic
def isand(elem, *args):
    for x in args:
        if not elem is x:
            return False
    return True


def isor(elem, *args):
    for x in args:
        if elem is x:
            return True
    return False


def eqand(elem, *args):
    for x in args:
        if not elem == x:
            return False
    return True


def eqor(elem, *args):
    for x in args:
        if elem == x:
            return True
    return False



from pyvorse.pipeline.datacell import dcell
from typing import Callable
from pyvorse.core import basis

class router(basis):
    """Class encapsulates functionality of running pipelines.
    """
    def __init__(self, func: Callable = None, meta = {}):
        if not callable(func):
            raise TypeError("Pointer not callable")
        self.pointer = func # can be None
        self.staticmeta = meta
        self.initial = None  # can be None
        self.final = None  # can be None
        self.prestep = None  # can be None
        self.poststep = None  # can be None

    def setup(self, func: Callable):
        self.pointer = func
        return func
    
    def initalizer(self, func: Callable):
        self.initial = func
        return func

    def finalizer(self, func: Callable):
        self.final = func
        return func
    
    def callBeforeStep(self, func: Callable):
        self.prestep = func
        return func
    
    def callAfterStep(self, func: Callable):
        self.poststep = func
        return func

    def launch(self, cell: dcell):
        """Run pipeline with box"""
        if callable(self.initial):
            self.initial(cell, self.staticmeta)
        #=============================
        self._launch(cell)
        #=============================
        if callable(self.final):
            self.final(cell, self.staticmeta)
        return self
    
    def _step(self, cell, selected):
        if callable(self.prestep):
            self.prestep(cell, self.staticmeta)
        #============================
        if not callable(selected):
            return
        selected(cell, self.staticmeta)
        #============================
        handler = self.pointer(cell, self.staticmeta)
        cell["tmp-handler"] = handler
        #============================
        if callable(self.poststep):
            self.poststep(cell, self.staticmeta)
        #============================
        del cell["tmp-handler"] # it`s only temporal metadata for poststep
        return handler
    
    
    def _launch(self, cell):
        if not callable(self.pointer):
            return
        handler = self.pointer(cell, self.staticmeta)
        if cell.stopped:
            return
        #============================
        while not cell.stopped and handler != None:
            handler = self._step(cell, handler)
    
    def __call__(self, cell: dcell):
        return self.launch(cell)
        



from pyvorse.data.simple import loop, scope, stack
from pyvorse.data.smart import smartlist, handle, objectAddress
from pyvorse.data.deep import deep, deepclass, dci
from pyvorse.data.table import Table, Column, Row
from pyvorse.data.spiders import spider



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyvorse',
 'pyvorse.core',
 'pyvorse.data',
 'pyvorse.mathematics',
 'pyvorse.pipeline',
 'pyvorse.randoms',
 'pyvorse.tests']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'pyvorse',
    'version': '1.0.4',
    'description': 'Simple essentials extension library of additions to vanilla python.',
    'long_description': '# pyvorse\n## Simple essentials extension library of additions to vanilla python.\n\n---\n\n### Includes:\n- [core]:\n  - @attribute - converts any function with (arg: str) to attribute processor\n  - w. - attribute processor to faster writing short words, dont use it in speed is important. For obvious reasons, such a word notation is guaranteed to conform to python\'s naming conventions. w.var_name = "var_name"\n  - class: basis - extended base class\n  - @throws(*exceptions_cls) - marks function with classes, that can be thrown\n  - isthrows(exception_cls) - test function to be marked with `@throws()`\n  - trycall(func, *args, **kwargs) -> tuple[result, exception] - catches exceptions in function call\n  - @safe - converts function to its copy, but automatically called inside `trycall()`, and returns tuple[result, exception]\n  - Operators - enum for @calloperators\n  - @calloperators(op: Operators = Operators.All) - converts function to object, that calls with secon operand and returns result of it function on using operator, selected in argument. Operand still can be called as function.\n  - reprint() - operand-variation of print function, can be called like `reprint - w.text_to_show`. If prints only one value, value will be returned as result.\n  - @functor - allows function to be called with self object as first argument. to work with its fields as example.\n  - class: unibox - class, that locks value inside field `.inbox`, and contains dict for metadata, and can be stopped with `.stop()`, and be test for it by `.stopped`, or `.isStopped()`\n  - copyAllProps(dest, source) - function copies all attributes from source to dest, but ignores attributes with name starting with duble underscore \'__\'\n- [data]:\n  - //ALL_FROM_SUBMODULES//\n  - [deep]\n    - deep - Implements a simplified version of javascript objects based on the "dict" class. Can be created directly or through a class of similar objects that extends "dci". Prototypal inheritance. All methods to look deeper starts with word «deep», like `.deepget(name)`. More details on the wiki.\n    - dci - empty base deepclass, use its `.extend(new_prototype: deep, *names: str)` method, that returns new deepclass object\n    - class: DeepCalss - *ignore it, nothing more then python`s class of «dci»*\n  - [positioned]\n    - class: poslist - list, based on dict. Aloows to contain skips inside.\n  - [simple]\n    - class: loop - list, but handles index overload like list is lopped\n    - class: scope - dict, but allows get-access to items like access to attributes, and attribute-like set access, but onky for keys, already present in dict. Allows to find item by number of a key (it automatically sorted in dicts). Allows getting values dynamic/ Value is a function to be called with self dict as first argument. `.getDynamic(key)` \n    - class: stack - list, extended only, no changes. `.setvalidation(validator: Callable)` - validates values on pushing to the stack. `.push(value)` and `.leftpush(value)`, `.pushes(value)` and `.leftpushes(value)`. Raises ValueError, if loses validation. `.pull()` and `.leftpull(value)` - works like default `.pop()` but without arguments, for last and first element.\n  - [smart]\n    - class: smartlist - list, but with additional methods. find, select, filtered (iter), filter, any, all, astuole, asslist, asset, foreach, mapper, reducer, sort, sliced, slicedstart, slicedend. Working with predicates and callbacks. Details on wiki.\n    - class: handle - Primitive handlers container, simplest pipeline. Objects can be used as decorators, or called with callback passed. On each step calls callback with box on first argument.\n    - class: objectAddress - Create, object `adr = objectAddress()`, write your way `adr["Configs"]["chances"]["spawn"](lambda pos: pos[SNEED])`, and use it on data structire `adr.use(CONFIGS_ROOT)`. It returns same for `CONFIGS_ROOT["Configs"]["chances"]["spawn"][SNEED]`, but can be used nultiple times and dynamically.\n  - [spiders]\n    - class: spider(seeker: Callable, worker: Callable) - spiders can work with complex data structures. Seeker selexts new node, worker does something here. Seeker and worker will be called with (struct, dot, box). Use `box.stop()` to finish spiders`s walk. But worker will be called even if you stop it in seeker. \n  - [table]\n    - class: Table - contains headers and rows. use `table[index_column][index_row]` or better `table.getitem(row:int, col:int)` for extracting values. Use `adddefaultheader(self, name: str, align: int = 0, default=None)` to add new column. You can use `.getdynamic(row:int, col:int)` to work with dynamic values. functions, that will be called with (table: Table, row:int, col:int). For example - auto-numerations. Rendering method not presented! Inherit class, and add any render you want, markdown or html for example.\n    - class: Column - saves names and align for headers. Saves default values for tables.\n    - class: Row: poslist - class represents rows in the table\n- [mathematics] - Whole default math, but with 2 new functions. `trin(n)` for triangular number and count of diagonals in any shape. `hexcount(n)` - for calculatinc number of hexagonal cells in radius from one.\n- [pipeline] - simple tools for complex piplines\n  - //ALL_FROM_SUBMODULES//\n  - [datacell]\n    - class: dcell: core.unibox - just container for value in pipeline.\n  - [mains] - Default pipline functions containers:\n    - DictionaryPipeline\n    - ListPipeline\n  - [routers] - full pipeline conveyor build. Cotains static metadata. Uses pointer to select new function, and it can be method from pipeline container.\n    - class: router - Read on wiki! Use `r.launch(cell)` or `r(cell)`. All functions in pipeline will be called with `(cell, staticmeta)`, with staticmeta from router.\n- [tests]\n  - //ALL_FROM_SUBMODULES//\n  - [exprs]\n    - iterable(o) - allows to get iterator from object, and using `o` inside for..in cycle\n    - nextable(o) - allows to use `next(o)`\n    - forable(o) - deeper version of `iterable(o)` test.\n    - instanceof(o, *cls) - alternative notation to isinstance, like `instanceof(my_num, int, float, complex)`\n    - withable(o) - allows with.. as..\n    - awithable(o) - allows async with.. as..\n    - isinrange(_list, index) - test index for a range in the index\n  - [other]\n    - hashable(o) - can be used as key for dictionary or not.\n    - haskey(o, key) - like hasattr, but tests for `o[key]`\n    - isand(o, *t) - tests o to be all of *t\n    - isor(o, *t) - tests o to be any of *t\n    - eqand(o, *t) - tests o to be equal to all of *t\n    - eqor(o, *t) - tests o to be equal to any of *t\n    - haslen(o) - can be item used in `len(o)` or not.',
    'author': 'Vladislav-Martian',
    'author_email': 'me.as.martian@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Vladislav-Martian/pyverse',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



__all__ = ["Stats"]

from pyvorse.core import basis
from typing import Callable, List, Any, Hashable
import json

class Stats(basis):
    "Methods to obtain statistics for data structures."

    @classmethod
    def statisticsJson(cls, data: List[Any]) -> str:
        "Returns json string with dictionary, with string-represented values as keys, and string with percentage as values."
        res = {}
        count = len(data)
        for v in data:
            key = repr(v)
            if not key in res:
                res[key] = 0
            else:
                res[key] += 1
        for key, value in res.items():
            res[key] = str(round(100 * (value / count), 2)) + '%'
        return json.dumps(res, indent=4, ensure_ascii=False)

    @classmethod
    def statisticsDict(cls, data: List[Hashable]) -> str:
        "Returns json string with dictionary, with string-represented values as keys, and string with percentage as values."
        res = {}
        count = len(data)
        for v in data:
            if not v in res:
                res[v] = 0
            else:
                res[v] += 1
        for key, value in res.items():
            res[key] = str(round(100 * (value / count), 2)) + '%'
        return res



__all__ = ["iterable", "nextable", "forable", "instanceof", "withable", "awithable", "isinrange"]

from typing import overload


def instanceof(obj, *clss):
    "Clone of isinstance() default function, but with inline notation"
    return isinstance(obj, tuple(clss))


def iterable(elem):
    """
    Returns True if element is iterable and vice-versa
    """
    try:
        iter(elem)
    except:
        return False
    else:
        return True


def nextable(elem):
    """
    Returns True if element is iterable and vice-versa
    """
    try:
        next(elem)
    except:
        return False
    else:
        return True


def forable(elem):
    """
    Returns True if element is able to use in for..in and vice-versa
    """
    try:
        x = iter(elem)
        next(x)
    except:
        return False
    else:
        return True


def withable(elem):
    """
    Returns True if element is able to use <with elem as var:> and vice-versa
    """
    return hasattr(elem, "__enter__") and hasattr(elem, "__exit__")


def awithable(elem):
    """
    Returns True if element is able to use <async with elem as var:> and vice-versa
    """
    return hasattr(elem, "__aenter__") and hasattr(elem, "__aexit__")


def isinrange(_list, index):
    "Tests is index in range of a list or tuple"
    try:
        _list[index]
        return True
    except IndexError:
        return False